/*
** ibench.c - a small benchmarking/stress-testing program for IDENT servers.
**
** Copyright (c) 1997-1999 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
**
**
** Please note that this program *should* be run on the same machine
** as the IDENT server you wish to test. It assumes it can connect to
** it via the loopback interface.
**
** If you specify a server address using the -rADDR option, then
** the lookups will be for the *servers* username (normally "nobody"),
** and not the username "ibench" runs as.
*/

#include "config.h"

#include <stdio.h>
#include <errno.h>
#include <signal.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <stdlib.h>
#include <time.h>
#include <pwd.h>
#include <string.h>
#include <sys/types.h>

#ifndef SYS_SOCKET_H_INCLUDED
#define SYS_SOCKET_H_INCLUDED
#include <sys/socket.h>
#endif

#include <netinet/in.h>
#include <arpa/inet.h>

#include "plib/petopt.h"
#include "pidentd.h"


int debug = 0;
int verbose = 0;

struct sockaddr_gen host_sin;
char *username = NULL;
int rev_flag = 0;
volatile int abort_test = 0;
int ntests = 0;
int ignore_reply = 0;
int test_len = 60;
int max_tests = 0;


static RETSIGTYPE
sigint_handler(int sig)
{
    abort_test = 1;
}

static RETSIGTYPE
sigalrm_handler(int sig)
{
    abort_test = 1;
}

/*
** Perform an IDENT lookup.
** Returns: -1 in case of a network or system error
**           0 if all was OK
**           1 if the reply was malformed
**           2 if the reply was with the wrong username
*/
static int
run_test(void)
{
    int fd, buflen, tlen;
    socklen_t len;
    struct sockaddr_gen our, sin;
    char buf[1024], buf2[1024];
    buflen = 0;
    

    ++ntests;
    if (max_tests > 0 && ntests > max_tests)
    {
	abort_test = 1;
	return 0;
    }
    
    sin = host_sin;

    fd = socket(SGFAM(sin), SOCK_STREAM, 0);
    if (fd < 0)
    {
	if (errno != EINTR)
	    perror("socket");
	return -1;
    }
    
    if (connect(fd, (struct sockaddr *) &sin, SGSOCKSIZE(sin)) < 0)
    {
	if (errno != EINTR)
	    perror("connect");
	goto Fail;
    }

    len = sizeof(our);
    if (getsockname(fd, (struct sockaddr *) &our, &len) < 0)
    {
	if (errno != EINTR)
	    perror("getsockname");
	goto Fail;
    }

#ifdef HAVE_SNPRINTF
    if (rev_flag)
	snprintf(buf, sizeof(buf),
		 "%d , %d\r\n",
		 ntohs(SGPORT(sin)),
		 ntohs(SGPORT(our)));
    else
	snprintf(buf, sizeof(buf),
		 "%d , %d\r\n",
		 ntohs(SGPORT(our)),
		 ntohs(SGPORT(sin)));
#else
    if (rev_flag)
	sprintf(buf, "%d , %d\r\n",
		ntohs(SGPORT(sin)),
		ntohs(SGPORT(our)));
    else
	sprintf(buf, "%d , %d\r\n",
		ntohs(SGPORT(our)),
		ntohs(SGPORT(sin)));
#endif
    
    if (write(fd, buf, strlen(buf)) < 0)
    {
	if (errno != EINTR)
	    perror("write");
	goto Fail;
    }

    if (shutdown(fd, 1) < 0)
    {
	if (errno != EINTR)
	    perror("shutdown");
	goto Fail;
    }

    tlen = 0;
    while (buflen < (sizeof(buf)-1) &&
	   (tlen = read(fd, buf+buflen, (sizeof(buf)-1)-buflen)) > 0)
    {
	buflen += tlen;
    }

    if (buflen == (sizeof(buf)-1))
    {
	fprintf(stderr, "Malformed reply (too long, %d bytes)\n",
		buflen);
	goto Fail;
    }

    if (tlen < 0)
    {
	if (errno != EINTR)
	    perror("read");
	goto Fail;
    }
    
    close(fd);
    
    buf[buflen] = '\0';

    buflen = sscanf(buf, " %*d , %*d : USERID : %*s :%[^\r\n]", buf2);

    if (buflen != 1)
    {
	fprintf(stderr, "Malformed reply (%d): %s\n", buflen, buf);
	return 1;
    }

    if (!ignore_reply && strcmp(buf2, username) != 0)
    {
	fprintf(stderr, "Incorrect username: %s != %s\n", buf2, username);
	return 2;
    }

    return 0;
    
  Fail:
    close(fd);
    
    if (errno == EINTR)
	return 0;
    
    return -1;
}

static PETOPTS pov[] =
{
    {
	'V',
	POF_NONE,
	"Version",
	NULL,
	"Print version"
    },
    
    {
	'h',
	POF_NONE,
	"HeLP",
	NULL,
	"Print help"
    },
    
    {
	'v',
	POF_BOOL|POF_OPT,
	"Verbose",
	&verbose,
	"Enable verbose mode"
    },
    
    {
	'd',
	POF_OPT|POF_INT,
	"DeBuG",
	&debug,
	"Enable debug mode"
    },
    
    {
	'i',
	POF_BOOL,
	"Ignore",
	&ignore_reply,
	"Ignore IDENT reply"
    },
    
    {
	'r',
	POF_STR,
	"Remote",
	NULL,
	"Remote mode"
    },
    
    {
	'p',
	POF_STR,
	"Port",
	NULL,
	"TCP/IP Port"
    },
    
    {
	'u',
	POF_STR,
	"Username",
	&username,
	"Expected username"
    },
    
    {
	't',
	POF_INT,
	"Timeout",
	&test_len,
	"Test timeout"
    },
    
    {
	'n',
	POF_INT,
	"Max-Tests",
	&max_tests,
	"Test limit"
    },
    
    { -1, 0, NULL, NULL, NULL }
};


static void
program_header(FILE *fp)
{
    fprintf(fp, "[Ibench, version %s (built %s %s)]\n",
	    server_version,
	    __DATE__, __TIME__);
}

void
usage(PETOPT *pp,
      FILE *fp)
{
    fputs("Usage: ibench [OPTIONS]\n", fp);
    fputs("\n", fp);
    petopt_print_usage(pp, fp);
}



static int
parse_option(PETOPT *pp,
	     PETOPTS *pov,
	     const char *arg)
{
    int port;
    
    
    switch (pov->s)
    {
      case 'h':
	usage(pp, stdout);
	exit(EXIT_SUCCESS);

      case 'V':
	program_header(stdout);
	exit(EXIT_SUCCESS);
	
      case 'r':
	rev_flag = 1;
	if (str2addr(arg, &host_sin) < 0)
	    return POE_INVALID;

	if (username == NULL)
	    username = "nobody";
	break;

      case 'p':
	if (str2port(arg, &port) < 0)
	    return POE_INVALID;
	SGPORT(host_sin) = htons(port);
	break;
	
      default:
	return 1;
    }

    return 0;
}

int
main(int argc,
     char *argv[])
{
    int len, ecode;
    time_t start, stop;
    struct passwd *pp;
    PETOPT *pop;

    
    ecode = petopt_setup(&pop, 0,
			 argc, argv,
			 pov,
			 parse_option, NULL);
    if (ecode)
    {
	if (ecode > 0)
	    fprintf(stderr, "%s: petopt_parse(): %s\n",
		    argv[0], strerror(ecode));
	exit(EXIT_FAILURE);
    }

    if (str2addr("127.0.0.1:113", &host_sin) < 0)
    {
	fprintf(stderr, "%s: can not build loopback address\n",
		argv[0]);
	exit(EXIT_FAILURE);
    }
    
    username = NULL;

    ecode = petopt_parse(pop, &argc, &argv);
    if (ecode)
    {
	if (ecode > 0)
	    fprintf(stderr, "%s: petopt_parse(): %s\n",
		    argv[0], strerror(ecode));
	exit(EXIT_FAILURE);
    }
    
    if (username == NULL)
    {
	pp = getpwuid(getuid());
	if (pp == NULL)
	{
	    perror("getpwuid()");
	    exit(EXIT_FAILURE);
	}
	username = pp->pw_name;
    }
    
    ntests = 0;
    ecode = 0;
    
    s_signal(SIGINT, sigint_handler);
    s_signal(SIGALRM, sigalrm_handler);
    
    printf("Test started, will run for %d seconds (or press Ctrl-C to terminate).\n", test_len);

    alarm(test_len);
    time(&start);
    
    while (!abort_test && (ecode = run_test()) == 0)
	;
    time(&stop);

    len = stop-start;
    if (len == 0)
	len = 1;

    if (ecode == 0 || (ecode == -1 && abort_test))
    {
	printf("Test OK: %d requests in %d seconds (%d requests/s)\n",
	       ntests, len, ntests/len);
	exit(EXIT_SUCCESS);
    }
    else
    {
	printf("Test FAILED after %d requests in %d seconds (%d requests/s)\n",
	       ntests, len, ntests/len);
	exit(EXIT_FAILURE);
    }
}
