/*
** k_linux.c - Linux 0.99.13q or later kernel access functions
**
** Copyright (c) 1997-1998 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#include <stdio.h>
#include <syslog.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#ifdef DEBUG
#include <syslog.h>
#endif

#include "pidentd.h"

#ifdef HAVE_IPV6
#define FP_INET		0
#define FP_INET6	1
#endif


/*
** Make sure we are running on a supported OS version
*/
int
ka_init(void)
{
    return 0; /* We always succeed */
}


int
ka_open(void **misc)
{
#ifdef HAVE_IPV6
    static FILE *fp[2];
    int e6, e;

    while ((fp[FP_INET6] = fopen("/proc/net/tcp6", "r")) == NULL &&
	   errno == EINTR)
	;
    e6 = errno;
    
    while ((fp[FP_INET] = fopen("/proc/net/tcp", "r")) == NULL &&
	   errno == EINTR)
	;
    e = errno;
    
    if (fp[FP_INET] == NULL && fp[FP_INET6] == NULL)
    {
	syslog(LOG_ERR, "fopen(\"/proc/net/tcp6\", \"r\"): %s", strerror(e6));
	syslog(LOG_ERR, "fopen(\"/proc/net/tcp\", \"r\"): %s", strerror(e));
	return -1;
    }
#else
    FILE *fp;

    while ((fp = fopen("/proc/net/tcp", "r")) == NULL && errno == EINTR)
	;
    
    if (fp == NULL)
    {
	syslog(LOG_ERR, "fopen(\"/proc/net/tcp\", \"r\"): %m");
	return -1;
    }
#endif

    *misc = (void *) fp;
    return 0;
}


static int
ka_lookup4(FILE *fp,
	   struct kernel *kp,
	   unsigned long r_laddr,
	   unsigned long r_raddr)
{
    long dummy;
    char buf[512];
    unsigned long myladdr, myraddr;
    int r_lport, r_rport, mylport, myrport;
    int euid;

    if (fp == NULL)
	return -1;
    
    r_lport = ntohs(SGPORT(kp->local));
    r_rport = ntohs(SGPORT(kp->remote));

    kp->ruid = NO_UID;
    rewind(fp);

    /* eat header */
    while (fgets(buf, sizeof(buf)-1,fp) == NULL)
	;

    while (fgets(buf, sizeof(buf)-1, fp) != NULL)
    {
	if (sscanf(buf, "%ld: %lX:%x %lX:%x %lx %lX:%lX %lx:%lX %lx %d %ld %ld",
		   &dummy, &myladdr, &mylport, &myraddr, &myrport,
		   &dummy, &dummy, &dummy, &dummy, &dummy, &dummy,
		   &euid, &dummy, &dummy) >= 12)
	{
	    if (myladdr == r_laddr && mylport == r_lport &&
		myraddr == r_raddr && myrport == r_rport)
	    {
		kp->euid = euid;
		return 0;
	    }
	}
    }

    return -1;
}

static int
ka_lookup6(FILE *fp,
	   struct kernel *kp,
	   struct in6_addr *r_laddr,
	   struct in6_addr *r_raddr)
{
    long dummy;
    unsigned dummy2;
    char buf[512];
    struct in6_addr myladdr, myraddr;
    int r_lport, r_rport, mylport, myrport;
    int euid;
    
    if (fp == NULL)
	return -1;
    
    r_lport = ntohs(SGPORT(kp->local));
    r_rport = ntohs(SGPORT(kp->remote));

    kp->ruid = NO_UID;
    rewind(fp);
    
    /* eat header */
    while (fgets(buf, sizeof(buf)-1, fp) == NULL)
	;
    while (fgets(buf, sizeof(buf)-1, fp) != NULL)
    {
	if (sscanf(buf,"%ld: %8x%8x%8x%8x:%x %8x%8x%8x%8x:%x %lx %lX:%lX %x:%lX %lx %d %ld %ld",
		   &dummy, &myladdr.s6_addr32[0], &myladdr.s6_addr32[1],
		   &myladdr.s6_addr32[2], &myladdr.s6_addr32[3], &mylport,
		   &myraddr.s6_addr32[0], &myraddr.s6_addr32[1],
		   &myraddr.s6_addr32[2], &myraddr.s6_addr32[3], &myrport,
		   &dummy, &dummy, &dummy, &dummy2, &dummy, &dummy,
		   &euid, &dummy, &dummy) >= 18)
	{
	    if (!memcmp(&myladdr, r_laddr, sizeof(struct in6_addr)) &&
		mylport == r_lport &&
		!memcmp(&myraddr, r_raddr, sizeof(struct in6_addr)) &&
		myrport == r_rport)
	    {
		kp->euid = euid;
		return 0;
	    }
	}
    }
    
    return -1;
}

int
ka_lookup(void *vp, struct kernel *kp)
{
#ifdef HAVE_IPV6 
    struct in6_addr r_laddr, r_raddr;
    
    switch(SGFAM(kp->remote)) {
    case AF_INET:
	memset(&r_laddr, 0, sizeof(r_laddr));
	memset(&r_raddr, 0, sizeof(r_raddr));
	r_laddr.s6_addr16[5] = 0xffff;
	r_raddr.s6_addr16[5] = 0xffff;
	r_laddr.s6_addr32[3] = *(UINT32 *)SGADDRP(kp->local);
	r_raddr.s6_addr32[3] = *(UINT32 *)SGADDRP(kp->remote);
	break;
    case AF_INET6:
	memcpy(&r_laddr, SGADDRP(kp->local), sizeof(r_laddr));
	memcpy(&r_raddr, SGADDRP(kp->remote), sizeof(r_raddr));
	break;
    default:
	return -1;
    }
    
    if(!ka_lookup6(((FILE **)vp)[FP_INET6], kp, &r_laddr, &r_raddr))
	return 0;
    
    if(SGFAM(kp->remote) == AF_INET6 && !IN6_IS_ADDR_V4MAPPED(&r_raddr))
	return -1;

    return ka_lookup4(((FILE **)vp)[FP_INET],
			   kp,
			   r_laddr.s6_addr32[3],
			   r_raddr.s6_addr32[3]);

#else
    if (SGFAM(kp->remote) == AF_INET)
    {
	return ka_lookup4(((FILE **)vp)[FP_INET],
			   kp,
			   * (UINT32 *) SGADDRP(kp->local),
			   * (UINT32 *) SGADDRP(kp->remote));
    }
#endif

    return -1;
}
