/*
** main.c - Main entrypoint for Pidentd 3.0
**
** Copyright (c) 1997-1999 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <signal.h>
#include <grp.h>

#include "pidentd.h"
#include "plib/petopt.h"

#if defined(HAVE_PTHREAD_SETCONCURRENCY)
#  include <pthread.h>
extern int pthread_setconcurrency (int __level) __THROW;
#elif   defined(HAVE_THR_SETCONCURRENCY)
#  include <thread.h>
#endif

char *argv0 = "identd";

int debug = 0;
uid_t server_uid = NO_UID;
gid_t server_gid = ROOT_GID;
char *pidfile_path = PATH_PIDFILE;

int listen_sock = -1;
int listen_port = IPPORT_IDENT;
struct sockaddr_gen listen_addr;
int listen_backlog = 256;

int requests_max = 0;

#ifdef HAVE_LIBDES
int encrypt_flag = 0;
char *encrypt_keyfile = PATH_KEYFILE;
#endif

int socket_type = -1;
int init_mode = 0;

int enable_socks = 0;

pthread_attr_t cattr_detached;


static PETOPTS pov[] =
{
    {
	'h',
	POF_NONE,
	"HeLP",
	NULL,
	"Print help"
    },
    
    {
	'V',
	POF_NONE,
	"Version",
	NULL,
	"Print version"
    },
    
    {
	'd',
	POF_OPT|POF_INT,
	"DeBuG",
	&debug,
	"Enable debug mode"
    },
    
#ifdef HAVE_LIBDES
    {
	'E',
	POF_BOOL,
	"Encryption",
	&encrypt_flag,
	"Enable encryption"
    },
#endif
    
    {
	'n',
	POF_BOOL,
	"Uid-only",
	&uidonly_flag,
	"Return UID numbers"
    },
    
    {
	'e',
	POF_BOOL,
	"Extensions",
	&extensions_enabled,
	"Enable protocol extensions"
    },
    
    {
	'm',
	POF_BOOL,
	"Multiquery",
	&multiquery_enabled,
	"Enable multiquery mode"
    },
    
    {
	'w',
	POF_NONE,
	"inetd-WAIT-mode",
	NULL,
	"Start in inetd/wait mode"
    },
    
    {
	'i',
	POF_NONE,
	"Inetd-NOWAIT-mode",
	NULL,
	"Start in inetd/nowait mode"
    },
    
    {
	'I',
	POF_NONE,
	"Inittab-mode",
	NULL,
	"Start in inittab mode"
    },
    
    {
	'b',
	POF_NONE,
	"Daemon-mode",
	NULL,
	"Start in daemon background mode"
    },
    
    {
	'N',
	POF_BOOL,
	"Noident",
	&noident_flag,
	"Check for .noident files"
    },
    {
	'p',
	POF_STR,
	"Port",
	NULL,
	"TCP/IP listening port"
    },
    
    {
	'a',
	POF_STR,
	"Address",
	NULL,
	"TCP/IP listening address"
    },
    
    {
	't',
	POF_INT,
	"TimeOut",
	&request_timeout,
	"Request timeout"
    },
    
    {
	'g',
	POF_STR,
	"GRouP-ID",
	NULL,
	"Group"
    },
    
    {
	'u',
	POF_STR,
	"User-ID",
	NULL,
	"User"
    },
    
    {
	'L',
	POF_STR,
	"Syslog-facility",
	NULL,
	"Facility for system logging"
    },
    
    {
	'C',
	POF_STR,
	"ConFiG-file",
	NULL,
	"Configuration file"
    },
    
    {
	'P',
	POF_STR,
	"Pid-file",
	&pidfile_path,
	"Where to store the PID"
    },
    
    {
	'K',
	POF_INT,
	"Kernel-threads",
	&kernel_threads,
	"Max number of kernel threads"
    },
    
    {
	'o',
	POF_STR,
	"OPerating-SYStem",
	&opsys,
	"Operating system token"
    },

    {
	'Q',
	POF_BOOL,
	"Proxy-Queries",
	&allow_proxy_query,
	"Enable proxy queries"
    },

    {
	'S',
	POF_STR|POF_OPT,
	"Socks",
	NULL,
	"Enable SOCKS5 IDENT support"
    },
    
    { -1, 0, NULL, NULL, NULL }
};

static void
program_header(FILE *fp)
{
    fprintf(fp, "[Pidentd, version %s (compiled for %s) - %s %s]\n",
	    server_version, osinfo_build,
	    __DATE__, __TIME__);
}



static void
usage(PETOPT *pp,
      FILE *fp)
{
    fprintf(fp, "Usage: %s [options]\n", argv0);

    fputs("\n\
This daemon implements the IDENT protocol as specified by RFC1413.\n\
More information may be found at:\n\
\thttp://www.lysator.liu.se/~pen/pidentd\n\
\n\
Command line options:\n", fp);
    
    petopt_print_usage(pp, fp);
    
    fputs("\n\
Long options are case insensitive, and may be abbreviated either by\n\
truncation or by using only the uppercase characters. '[ARG]' denotes\n\
an optional argument and 'ARG' denotes a required argument.\n", fp);
}


static int
parse_option(PETOPT *pp,
	     PETOPTS *pov,
	     const char *arg)
{
    int code;

    
    switch (pov->s)
    {
      case 'S':
	socks_init(arg);
	enable_socks = 1;
	return 0;
	
      case 'o':
	if (arg == NULL)
	{
	    opsys = "OTHER";
	    return 0;
	}
	break;
	
      case 'h':
	usage(pp, stdout);
	exit(EXIT_SUCCESS);
	    
      case 'w':
	listen_sock = STDIN_FILENO;
	socket_type = SOCKTYPE_LISTEN;
	break;
	
      case 'i':
	listen_sock = STDIN_FILENO;
	socket_type = SOCKTYPE_CONNECTED;
	break;
	
      case 'I':
	listen_sock = -1;
	socket_type = SOCKTYPE_NOTSOCKET;
	init_mode = 1;
	break;
	
      case 'b':
	listen_sock = -1;
	socket_type = SOCKTYPE_NOTSOCKET;
	break;
	
      case 'p':
	if (str2port(arg, &listen_port) < 0)
	    return POE_INVALID;
	break;
	
      case 'a':
	if (str2addr(arg, &listen_addr) < 0)
	    return POE_INVALID;
	if (SGPORT(listen_addr) != 0)
	    listen_port = ntohs(SGPORT(listen_addr));
	break;
	
      case 'g':
	if (str2gid(arg, &server_gid) < 0)
	    return POE_INVALID;
	break;
	
      case 'u':
	if (str2uid(arg, &server_uid, &server_gid) < 0)
	    return POE_INVALID;
	break;

      case 'L':
	code = syslog_str2fac(arg);
	if (code < 0)
	    return POE_INVALID;
	closelog();
	s_openlog(argv0, LOG_PID|LOG_ODELAY, code);
	break;
	
      case 'C':
	if (conf_parse(arg, 0) < 0)
	{
	    if (socket_type == -1 || socket_type == SOCKTYPE_NOTSOCKET)
		fprintf(stderr, "%s: error parsing config file: %s\n",
			argv0, arg);
	    exit(EXIT_FAILURE);
	}
	break;
	
      case 'V':
	program_header(stdout);
	exit(EXIT_SUCCESS);
	
      default:
	return 1;
    }

    return 0;
}


void
drop_root_privs(void)
{
    if (server_uid == NO_UID)
    {
	if (str2uid("nobody", &server_uid, &server_gid) < 0)
	    server_uid = ROOT_UID;
    }

    if (setgroups(1, &server_gid) < 0)
	goto Error;
    
    if (server_gid != ROOT_GID)
	if (setgid(server_gid) < 0)
	    goto Error;
    
    if (server_uid != ROOT_UID)
	if (setuid(server_uid) < 0)
	    goto Error;

    return;
  Error:
    syslog(LOG_ERR, "Error while changing user/group privileges");
    exit(EXIT_FAILURE);
}    


int
main(int argc,
     char *argv[])
{
    int err;
    char *optarg;
    PETOPT *pop;
    int code = LOG_DAEMON;
    SERVER *servp = NULL;
    

    a_init();
    
    SGINIT(listen_addr);
    SGPORT(listen_addr) = htons(113);
    
    if ((optarg = getenv("IDENTD_DEBUG")) != NULL)
    {
	if (*optarg)
	    debug = atoi(optarg);
	else
	    ++debug;
    }

    if (argv[0] != NULL)
    {
	char *cp;
	
	cp = strrchr(argv[0], '/');
	if (cp != NULL)
	    argv0 = A_STRDUP(cp+1);
	else
	    argv0 = A_STRDUP(argv[0]);
    }
    
#ifdef SIGTTOU    
    s_signal(SIGTTOU, SIG_IGN);
#endif
    s_signal(SIGPIPE, SIG_IGN);
    
    s_openlog(argv0, LOG_PID|LOG_ODELAY, code);

    err = petopt_setup(&pop, POF_SYSLOG,
		       argc, argv,
		       pov,
		       parse_option, NULL);
    if (err)
    {
	if (err > 0)
	    fprintf(stderr, "%s: petopt_setup: %s\n",
		    argv[0], strerror(err));
	return EXIT_FAILURE;
    }
    
    /*
    ** Try to autodetect how we was started.
    */
    socket_type = socktype(STDIN_FILENO);
    
    if (debug)
	fprintf(stderr, "socktype = %d\n", socket_type);

    if (socket_type == SOCKTYPE_LISTEN || socket_type == SOCKTYPE_CONNECTED)
	listen_sock = STDIN_FILENO;

    conf_parse(PATH_ETC_CFGFILE, 1); /* /etc/identd.conf */
    if (s_strcasecmp(PATH_ETC_CFGFILE, PATH_CFGFILE) != 0)
	conf_parse(PATH_CFGFILE, 1);     /* /usr/local/etc/identd.conf */

    err = petopt_parse(pop, &argc, &argv);
    if (err)
    {
	if (err > 0)
	    fprintf(stderr, "%s: petopt_parse() failed: %s\n",
		    argv[0], strerror(err));
        return EXIT_FAILURE;
    }
    
    if (debug)
	program_header(stderr);

    if (socket_type == -1)
    {
	syslog(LOG_ERR, "unable to autodetect socket type");
	fprintf(stderr, "%s: unable to autodetect socket type\n",
		argv0);
	return EXIT_FAILURE;
    }
    
    if (ka_init())
    {
	syslog(LOG_ERR, "OS version mismatch - compiled for %s", osinfo_build);
	if (socket_type == SOCKTYPE_NOTSOCKET)
	    fprintf(stderr,
		    "%s: OS version mismatch - compiled for: %s\n",
		    argv[0], osinfo_build);
	return EXIT_FAILURE;
    }

    
    if (!debug && 
	getppid() != INIT_PID && !init_mode &&
	socket_type != SOCKTYPE_CONNECTED &&
	listen_sock < 0)
    {
	become_daemon();
    }

#ifdef HAVE_PTHREAD_SETCONCURRENCY
    pthread_setconcurrency(kernel_threads+8);
#else
#ifdef HAVE_THR_SETCONCURRENCY
    thr_setconcurrency(kernel_threads+8);
#endif
#endif
    syslog(LOG_DEBUG, "started");


    pthread_attr_init(&cattr_detached);
    pthread_attr_setdetachstate(&cattr_detached, PTHREAD_CREATE_DETACHED);
    
    
    if (socket_type != SOCKTYPE_CONNECTED &&
	!debug && pidfile_path != NULL)
    {
	pidfile_create(pidfile_path);
    }

    if (listen_port > 0)
	SGPORT(listen_addr) = htons(listen_port);
    
    timeout_init();
    
    servp = server_init(listen_sock,
			&listen_addr,
			(socket_type != SOCKTYPE_CONNECTED ?
			 listen_backlog : -1),
			requests_max,
			request_handler);
    if (servp == NULL)
    {
	if (debug)
	    fprintf(stderr, "%s: failed binding to the TCP/IP socket\n",
		    argv[0]);
	goto Exit;
    }


#ifdef HAVE_LIBDES
    if (encrypt_flag)
    {
	if (pdes_init(encrypt_keyfile) < 0)
	{
	    syslog(LOG_ERR, "encryption could not be initalized: %m");
	    if (debug)
	    {
		fprintf(stderr, "%s: encryption could not be initialized: ",
			argv[0]);
		perror("");
	    }
	    goto Exit;
	}
    }
#endif

#ifndef HAVE_THREADS
    /* When not using threads we must open the kernel device(s) in the
       request handling subprocess, and then we can't drop our root
       privs at this early stage (unless we are started from Inetd/nowait */
    if (socket_type != SOCKTYPE_CONNECTED)
    {
#endif

/* Sigh - stupid Linux handles threads like... Anyway, we'll have to
   add this kludge to work around the fact that threads in Linux can
   have different uid's... Luckily Linux doesn't need root to get at
   the needed information anyway. */
#ifdef __linux__
	drop_root_privs();
#endif
	if (kernel_init() < 0)
	{
	    if (debug)
		fprintf(stderr, "%s: failed opening kernel devices\n",
			argv[0]);
	    goto Exit;
	}

#ifndef __linux__
	drop_root_privs();
#endif
    
#ifndef HAVE_THREADS
    }
#endif
    
    if (socket_type != SOCKTYPE_CONNECTED)
    {
	if (debug)
	    fprintf(stderr, "entering server main loop\n");
    
	server_run_all(servp);
    }
    else
	return server_run_one(servp, 0, 1);

  Exit:
    syslog(LOG_DEBUG, "terminating");
    return EXIT_FAILURE;
}
