/*
** send.c - Format and send IDENT protocol replies.
**
** Copyright (c) 1997 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#include <stdio.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <string.h>
#include <syslog.h>
#include <pwd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "pidentd.h"

int uidonly_flag = 0;
int noident_flag = 0;
int result_syslog_level = -1;
int unknown_error_flag = 0;

char *charset = NULL;
char *opsys = "UNIX";


static int
check_noident(char *dir)
{
    char buf[2048];

    
    if (dir == NULL)
    {
	syslog(LOG_ERR, "check_noident: NULL directory");
	return -1;
    }
    
    if (strlcpy(buf, dir, sizeof(buf)) >= sizeof(buf))
    {
	syslog(LOG_ERR, "check_noident: buffer too small for dir: %s", dir);
	return -1;
    }

    if (strlcat(buf, "/.noident", sizeof(buf)) >= sizeof(buf))
    {
	syslog(LOG_ERR, "check_noident: buffer too small for dir: %s", dir);
	return -1;
    }

    if (access(buf, F_OK) == 0)
	return 1;
    else
	return 0;
}


void
send_error(FDBUF *fp,
	   CLIENT *cp,
	   const char *req,
	   const char *what)
{
    char buf[36];		/* 36 = max (24, 11 + longest 'what'(=25))*/
    char nab[64];

    
    if (unknown_error_flag || (what[0] == 'X' && what[1] == '-' && !debug))
	s_snprintf(buf, sizeof(buf), " : ERROR : UNKNOWN-ERROR");
    else
	s_snprintf(buf, sizeof(buf), " : ERROR : %s", what);

    fd_puts(fp, req);
    fd_puts(fp, buf);
    fd_putc(fp, '\n');
    
    if (result_syslog_level > -1)
	syslog(result_syslog_level,
	       "reply to %s: %s%s",
	       s_inet_ntox(&cp->rsin, nab, sizeof(nab)), req,
	       buf);
}


void
send_result(FDBUF *fp,
	    CLIENT *cp,
	    const char *req,
	    struct kernel *kp)
{
    char buf[2048], pbuf[2048];
    char nab[64];
    struct passwd pwb, *pp = NULL;
    uid_t uid;


    if (debug)
	fprintf(stderr, "send_result(%d) - ruid = %ld, euid = %ld\n",
		fp->fd, (long) kp->ruid, (long) kp->euid);

    /* XXX: This is no good, but.. */
    if (kp->ruid == NO_UID)
	uid = kp->euid;
    else
	uid = kp->ruid;

    if (!uidonly_flag || noident_flag)
	(void) s_getpwuid_r(uid, &pwb, pbuf, sizeof(pbuf), &pp);

    if (noident_flag && pp != NULL && check_noident(pp->pw_dir) == 1)
    {
	syslog(LOG_INFO, "User %s elected to use .noident", pp->pw_name);
	
	send_error(fp,
		   cp,
		   buf,
		   "HIDDEN-USER");
	return;
    }

#ifdef HAVE_LIBDES
    if (encrypt_flag)
    {
	char buffer[33];

	pdes_encrypt(kp, buffer);
	s_snprintf(buf, sizeof(buf),
		 " : USERID : OTHER%s%s :[%s]",
		  charset != NULL ? " , " : "",
		  charset != NULL ? charset : "",
		  buffer);
    }
    else
#endif
	
    if (!uidonly_flag && pp != NULL && strlen(pp->pw_name) < sizeof(buf)-128)
    {
	s_snprintf(buf, sizeof(buf),
		   " : USERID : %s%s%s :%s",
		   opsys,
		   charset != NULL ? " , " : "",
		   charset != NULL ? charset : "",
		   pp->pw_name);
    }
    else
    {
	s_snprintf(buf, sizeof(buf),
		   " : USERID : OTHER :%lu",
		   (unsigned long) uid);
    }

    fd_puts(fp, req);
    fd_puts(fp, buf);
    fd_putc(fp, '\n');
    
    if (result_syslog_level > -1)
	syslog(result_syslog_level,
	       "reply to %s: %s : %s",
	       s_inet_ntox(&cp->rsin, nab, sizeof(nab)),
	       req, buf);
}



void
send_version(FDBUF *fp,
	     CLIENT *cp)
{
    char buf[1024];
    char nab[64];
    
    s_snprintf(buf, sizeof(buf),
	       "0 , 0 : X-VERSION : pidentd %s for %s (%s %s)",
	       server_version,
	       osinfo_build, __DATE__, __TIME__);

    fd_puts(fp, buf);
    fd_putc(fp, '\n');
    
    if (result_syslog_level > -1)
	syslog(result_syslog_level,
	       "reply to %s: %s",
	       s_inet_ntox(&cp->rsin, nab, sizeof(nab)), buf);
}
