/*
** socks.c - SOCKS support
**
** Copyright (c) 1999 Peter Eriksson <pen@lysator.liu.se>
**
** This program is free software; you can redistribute it and/or
** modify it as you wish - as long as you don't claim that you wrote
** it.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
*/

#include "config.h"

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <string.h>

#include <sys/stat.h>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FLOCK
#include <sys/file.h>
#else
#include <fcntl.h>
#endif

#ifdef HAVE_LIBDB
#include <db.h>
#endif

#include "pidentd.h"

#include "socks.h"


int use_socks_db = 0;

static char *socks_srvidt_file = NULL;

#ifdef HAVE_LIBDB
static DB *db_h = NULL;
#endif


#ifdef HAVE_LIBDB
static int
socks_db_init(const char *srvidt_db)
{
    char buf[1024], *cp;
    DB_ENV *dbenv;
    int err;
    
    
    if (strlcpy(buf, srvidt_db, sizeof(buf)) >= sizeof(buf))
    {
	syslog(LOG_ERR, "socks_db_init: strlcpy: buffer too small");
	return -1;
    }
    
    cp = strrchr(buf, '.');
    if (!cp || strcmp(cp, ".db") != 0)
	if (strlcat(buf, ".db", sizeof(buf)) >= sizeof(buf))
	{
	    syslog(LOG_ERR, "socks_db_init: strlcat: buffer too small");
	    return -1;
	}
    
    A_NEW(dbenv);
    
    err = db_appinit(NULL, NULL, dbenv,
#ifdef HAVE_THREADS
		     DB_THREAD|
#endif
		     DB_CREATE|DB_INIT_LOCK|DB_INIT_MPOOL);
    
    if (err)
    {
	syslog(LOG_ERR, "socks_db_init: db_appinit: %s",
	       err > 0 ? strerror(err) : "<internal error>");
	a_free(dbenv);
	return -1;
    }
    
    /* We create it if it doesn't exist */
    err = db_open(buf, DB_HASH, DB_CREATE, 0600, dbenv, NULL, &db_h);
    if (err)
    {
	syslog(LOG_ERR, "socks_db_init: db_open: %s",
	       err > 0 ? strerror(err) : "<internal error>");
	db_h = NULL;
	a_free(dbenv);
	return -1;
    }

    return 0;
}
#endif


void
socks_init(const char *srvidt_file)
{
    if (srvidt_file == NULL)
	srvidt_file = getenv("SOCKS5_IDENTFILE");

#ifdef HAVE_LIBDB
    if (use_socks_db && socks_db_init(srvidt_file) == 0)
	return;
#endif
    
    socks_srvidt_file = A_STRDUP(srvidt_file);
}


#ifdef HAVE_LIBDB
static char *
socks_db_lookup(const char *idt,
		char *ubuf,
		size_t ulen)
{
    DBT key, data;
    int err;

    
    if (db_h == NULL)
	return NULL;
    
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));
    
    key.data = (char *) idt;
    key.size = strlen(idt);
    
    err = db_h->get(db_h, NULL, &key, &data, 0);
    if (err)
	return NULL; /* XXX: Should perhaps syslog some errors */
    
    if (ulen < data.size+1)
	return NULL; /* XXX: Should syslog */
    
    memcpy(ubuf, data.data, data.size);
    ubuf[data.size] = 0;
    return ubuf;
}
#endif


char *
socks_lookup(struct kernel *kp,
	     char *ubuf,
	     size_t ulen)
{
    int fd = -1, len;
    char ra[16], la[16];
    char idt[64], buf[256];
    char *cp;
    FDBUF *fb = NULL;
    
#ifndef HAVE_FLOCK
    struct flock fl;

    fl.l_type = F_RDLCK;
    fl.l_whence = SEEK_SET;
    fl.l_start = 0;
    fl.l_len = 0;
#endif

    /* XXX: Fixme: The "idt" generated is not compatible! IPv6 & ports
       due to the change in s_inet_ntox! */
      
    len = s_snprintf(idt, sizeof(idt),
		     "%s,%s",
		     s_inet_ntox(&kp->remote, ra, sizeof(ra)),
		     s_inet_ntox(&kp->local, la, sizeof(la)));
    if (len < 0)
	return NULL;

    if (debug)
	fprintf(stderr, "socks_lookup, checking for: %s\n",
		idt);

#ifdef HAVE_LIBDB
    if (socks_db_lookup(idt, ubuf, ulen) == 0)
	return ubuf;
#endif

    if (socks_srvidt_file == NULL)
	return NULL;
    
    fd = s_open(socks_srvidt_file, O_RDONLY);
    if (fd < 0)
	return NULL;

#ifndef HAVE_FLOCK
    fcntl(fd, F_SETLKW, &fl);
#else
    flock(fd, LOCK_EX);
#endif

    fb = fd_create(fd, 0);
	
    while (fd_gets(fb, buf, sizeof(buf)) >= 0)
    {
	if (debug)
	    fprintf(stderr, "sock_lookup, got: %s\n", buf);

	cp = strrchr(buf, ',');
	if (cp == NULL)
	{
	    syslog(LOG_ERR, "Invalid SOCKS srvidt file");
	    goto Fail;
	}

	*cp++ = '\0';
	
	if (strcmp(buf, idt) == 0)
	{
	    if (strlcpy(ubuf, cp, ulen) >= ulen)
	    {
		syslog(LOG_ERR,
		       "socks_lookup(): Username buffer too small");
		goto Fail;
	    }
	    
	    fd_destroy(fb);
	    s_close(fd);
	    return ubuf;
	}
    }

  Fail:
    fd_destroy(fb);
    s_close(fd);
    return NULL;
}
